from chainlit.config import config

from easierlit import EasierlitApp, EasierlitClient, EasierlitPersistenceConfig
from easierlit.chainlit_entry import (
    _apply_runtime_configuration,
    _should_register_default_data_layer,
)
from easierlit.runtime import get_runtime


def _clear_chainlit_hooks() -> None:
    config.code.password_auth_callback = None
    config.code.data_layer = None


def test_default_sqlite_data_layer_is_registered_when_no_external_db(monkeypatch, tmp_path):
    runtime = get_runtime()
    runtime.unbind()
    _clear_chainlit_hooks()
    monkeypatch.setattr("easierlit.chainlit_entry._CONFIG_APPLIED", False)

    monkeypatch.delenv("DATABASE_URL", raising=False)
    monkeypatch.delenv("LITERAL_API_KEY", raising=False)

    db_path = tmp_path / "default-sidebar.db"
    runtime.bind(
        client=EasierlitClient(run_func=lambda _app: None),
        app=EasierlitApp(),
        persistence=EasierlitPersistenceConfig(enabled=True, sqlite_path=str(db_path)),
    )

    _apply_runtime_configuration()

    assert config.code.data_layer is not None
    assert db_path.exists()
    assert config.ui.cot == "full"

    runtime.unbind()
    _clear_chainlit_hooks()


def test_default_sqlite_is_not_registered_when_database_url_exists(monkeypatch):
    runtime = get_runtime()
    runtime.unbind()
    _clear_chainlit_hooks()
    monkeypatch.setattr("easierlit.chainlit_entry._CONFIG_APPLIED", False)

    monkeypatch.setenv("DATABASE_URL", "postgresql://example")
    monkeypatch.delenv("LITERAL_API_KEY", raising=False)

    runtime.bind(
        client=EasierlitClient(run_func=lambda _app: None),
        app=EasierlitApp(),
        persistence=EasierlitPersistenceConfig(enabled=True),
    )

    assert _should_register_default_data_layer() is False
    _apply_runtime_configuration()
    assert config.code.data_layer is None
    assert config.ui.cot == "full"

    runtime.unbind()
    _clear_chainlit_hooks()



import asyncio

from easierlit.models import OutgoingCommand
from easierlit.runtime import get_runtime


class _FakeDataLayer:
    def __init__(self):
        self.created_steps = []

    async def create_step(self, step_dict):
        self.created_steps.append(step_dict)


def test_apply_outgoing_command_initializes_http_context(monkeypatch):
    runtime = get_runtime()
    runtime.unbind()

    data_layer = _FakeDataLayer()
    context_calls = []

    def fake_init_http_context(*, thread_id: str, client_type: str):
        context_calls.append((thread_id, client_type))

    monkeypatch.setattr("easierlit.runtime.get_data_layer", lambda: data_layer)
    monkeypatch.setattr("easierlit.runtime.init_http_context", fake_init_http_context)

    command = OutgoingCommand(
        command="add_message",
        thread_id="thread-ctx",
        message_id="msg-ctx",
        content="hello",
        author="Bot",
        metadata={},
    )

    asyncio.run(runtime.apply_outgoing_command(command))

    assert context_calls == [("thread-ctx", "webapp")]
    assert data_layer.created_steps[0]["id"] == "msg-ctx"



import asyncio

import pytest

from easierlit import EasierlitApp, EasierlitClient
from easierlit.runtime import get_runtime


@pytest.fixture(autouse=True)
def _reset_runtime():
    runtime = get_runtime()
    runtime.unbind()
    yield
    runtime.unbind()


def test_register_and_unregister_session_mapping():
    runtime = get_runtime()
    runtime.register_session(thread_id="thread-1", session_id="session-1")
    assert runtime.get_session_id_for_thread("thread-1") == "session-1"

    runtime.unregister_session("session-1")
    assert runtime.get_session_id_for_thread("thread-1") is None


def test_dispatcher_consumes_outgoing_queue(monkeypatch):
    runtime = get_runtime()
    app = EasierlitApp()
    client = EasierlitClient(lambda _app: None)
    runtime.bind(client=client, app=app)

    async def scenario():
        runtime.set_main_loop(asyncio.get_running_loop())
        seen = []

        async def fake_apply(command):
            seen.append(command)

        monkeypatch.setattr(runtime, "apply_outgoing_command", fake_apply)

        await runtime.start_dispatcher()
        app.add_message(thread_id="thread-1", content="hello")
        await asyncio.sleep(0.3)
        await runtime.stop_dispatcher()

        assert len(seen) >= 1
        assert seen[0].command == "add_message"
        assert seen[0].thread_id == "thread-1"

    asyncio.run(scenario())



from .app import EasierlitApp
from .client import EasierlitClient
from .errors import (
    AppClosedError,
    DataPersistenceNotEnabledError,
    EasierlitError,
    RunFuncExecutionError,
    ThreadSessionNotActiveError,
    WorkerAlreadyRunningError,
    WorkerNotRunningError,
)
from .models import IncomingMessage, OutgoingCommand
from .server import EasierlitServer
from .settings import EasierlitAuthConfig, EasierlitPersistenceConfig

__all__ = [
    "AppClosedError",
    "DataPersistenceNotEnabledError",
    "EasierlitApp",
    "EasierlitAuthConfig",
    "EasierlitClient",
    "EasierlitError",
    "EasierlitPersistenceConfig",
    "EasierlitServer",
    "IncomingMessage",
    "OutgoingCommand",
    "RunFuncExecutionError",
    "ThreadSessionNotActiveError",
    "WorkerAlreadyRunningError",
    "WorkerNotRunningError",
]



import asyncio

import pytest
from chainlit.config import config

from easierlit import EasierlitApp, EasierlitAuthConfig, EasierlitClient
from easierlit.chainlit_entry import _apply_runtime_configuration
from easierlit.runtime import get_runtime


@pytest.fixture(autouse=True)
def _reset_state(monkeypatch):
    runtime = get_runtime()
    runtime.unbind()
    monkeypatch.setattr("easierlit.chainlit_entry._CONFIG_APPLIED", False)

    config.code.password_auth_callback = None
    config.code.data_layer = None

    yield

    runtime.unbind()
    monkeypatch.setattr("easierlit.chainlit_entry._CONFIG_APPLIED", False)
    config.code.password_auth_callback = None
    config.code.data_layer = None


def test_auth_config_sets_password_callback(monkeypatch):
    monkeypatch.delenv("DATABASE_URL", raising=False)
    monkeypatch.delenv("LITERAL_API_KEY", raising=False)

    auth = EasierlitAuthConfig(
        username="admin",
        password="admin",
        metadata={"role": "admin"},
    )
    runtime = get_runtime()
    runtime.bind(
        client=EasierlitClient(run_func=lambda _app: None),
        app=EasierlitApp(),
        auth=auth,
    )

    _apply_runtime_configuration()
    assert config.code.password_auth_callback is not None

    result = asyncio.run(config.code.password_auth_callback("admin", "admin"))
    assert result is not None
    assert result.identifier == "admin"
    assert result.metadata == {"role": "admin"}

    invalid = asyncio.run(config.code.password_auth_callback("admin", "wrong"))
    assert invalid is None


def test_auth_config_uses_custom_identifier(monkeypatch):
    monkeypatch.delenv("DATABASE_URL", raising=False)
    monkeypatch.delenv("LITERAL_API_KEY", raising=False)

    auth = EasierlitAuthConfig(
        username="admin",
        password="admin",
        identifier="owner",
    )
    runtime = get_runtime()
    runtime.bind(
        client=EasierlitClient(run_func=lambda _app: None),
        app=EasierlitApp(),
        auth=auth,
    )

    _apply_runtime_configuration()
    result = asyncio.run(config.code.password_auth_callback("admin", "admin"))
    assert result is not None
    assert result.identifier == "owner"


def test_auth_config_rejects_empty_credentials():
    with pytest.raises(ValueError):
        EasierlitAuthConfig(username="", password="admin")

    with pytest.raises(ValueError):
        EasierlitAuthConfig(username="admin", password="   ")


def test_auth_config_does_not_accept_jwt_secret_argument():
    with pytest.raises(TypeError):
        EasierlitAuthConfig(username="admin", password="admin", jwt_secret="x")  # type: ignore[call-arg]



from __future__ import annotations

import asyncio
import logging
import os
import secrets
import sys
from pathlib import Path

import chainlit as cl
from chainlit.auth import require_login
from chainlit.config import config
from chainlit.data import get_data_layer
from chainlit.user import User

# When this module is loaded by Chainlit's load_module(file_path), ensure the
# src root is importable so absolute imports keep working.
_THIS_FILE = Path(__file__).resolve()
_SRC_ROOT = _THIS_FILE.parents[1]
if str(_SRC_ROOT) not in sys.path:
    sys.path.insert(0, str(_SRC_ROOT))

from easierlit.errors import AppClosedError, RunFuncExecutionError
from easierlit.models import IncomingMessage
from easierlit.runtime import get_runtime
from easierlit.settings import EasierlitPersistenceConfig
from easierlit.sqlite_bootstrap import ensure_sqlite_schema

LOGGER = logging.getLogger(__name__)
RUNTIME = get_runtime()
_CONFIG_APPLIED = False
_APP_CLOSED_WARNING_EMITTED = False
_WORKER_FAILURE_UI_NOTIFIED = False


def _summarize_worker_error(traceback_text: str) -> str:
    lines = [line.strip() for line in traceback_text.strip().splitlines() if line.strip()]
    if lines:
        return lines[-1]
    return "Unknown run_func error"


def _apply_runtime_configuration() -> None:
    global _CONFIG_APPLIED
    if _CONFIG_APPLIED:
        return

    _apply_auth_configuration()
    _register_default_data_layer_if_needed()

    # Easierlit policy.
    config.ui.default_sidebar_state = "open"
    config.ui.cot = "full"
    _CONFIG_APPLIED = True


def _apply_auth_configuration() -> None:
    auth = RUNTIME.get_auth()
    if auth is None:
        return

    expected_username = auth.username
    expected_password = auth.password
    resolved_identifier = auth.identifier or auth.username
    resolved_metadata = auth.metadata or {}

    async def _password_auth_callback(username: str, password: str) -> User | None:
        if not secrets.compare_digest(username, expected_username):
            return None
        if not secrets.compare_digest(password, expected_password):
            return None
        return User(identifier=resolved_identifier, metadata=dict(resolved_metadata))

    cl.password_auth_callback(_password_auth_callback)


def _should_register_default_data_layer() -> bool:
    persistence = RUNTIME.get_persistence() or EasierlitPersistenceConfig()
    if not persistence.enabled:
        return False
    if config.code.data_layer is not None:
        return False
    if os.getenv("DATABASE_URL"):
        return False
    if os.getenv("LITERAL_API_KEY"):
        return False
    return True


def _register_default_data_layer_if_needed() -> None:
    if not _should_register_default_data_layer():
        return

    persistence = RUNTIME.get_persistence() or EasierlitPersistenceConfig()
    db_path = ensure_sqlite_schema(persistence.sqlite_path).resolve()
    conninfo = f"sqlite+aiosqlite:///{db_path}"

    @cl.data_layer
    def _easierlit_default_data_layer():
        from chainlit.data.sql_alchemy import SQLAlchemyDataLayer

        return SQLAlchemyDataLayer(conninfo=conninfo)

    LOGGER.info("Easierlit default SQLite data layer enabled at %s", db_path)


@cl.on_app_startup
async def _on_app_startup() -> None:
    global _APP_CLOSED_WARNING_EMITTED, _WORKER_FAILURE_UI_NOTIFIED
    _apply_runtime_configuration()
    _APP_CLOSED_WARNING_EMITTED = False
    _WORKER_FAILURE_UI_NOTIFIED = False
    RUNTIME.set_main_loop(asyncio.get_running_loop())
    await RUNTIME.start_dispatcher()

    try:
        data_layer = get_data_layer()
    except Exception:
        LOGGER.exception("Failed to initialize Chainlit data layer at startup.")
        data_layer = None

    if not require_login():
        LOGGER.warning(
            "Thread History sidebar is hidden by Chainlit policy because "
            "requireLogin=False. Configure Easierlit auth to enable it."
        )
    if data_layer is None:
        LOGGER.warning(
            "Thread History sidebar may be hidden because dataPersistence=False. "
            "Configure a data layer (or keep Easierlit default persistence enabled)."
        )


@cl.on_app_shutdown
async def _on_app_shutdown() -> None:
    global _APP_CLOSED_WARNING_EMITTED, _CONFIG_APPLIED, _WORKER_FAILURE_UI_NOTIFIED
    await RUNTIME.stop_dispatcher()

    client = RUNTIME.get_client()
    if client is None:
        return

    try:
        client.stop()
    except RunFuncExecutionError as exc:
        worker_error = client.peek_worker_error()
        summary = _summarize_worker_error(worker_error or str(exc))
        LOGGER.warning("run_func crash acknowledged during shutdown: %s", summary)
    finally:
        _CONFIG_APPLIED = False
        _APP_CLOSED_WARNING_EMITTED = False
        _WORKER_FAILURE_UI_NOTIFIED = False


@cl.on_chat_start
async def _on_chat_start() -> None:
    session = cl.context.session
    RUNTIME.register_session(thread_id=session.thread_id, session_id=session.id)


@cl.on_chat_resume
async def _on_chat_resume(_thread: dict) -> None:
    session = cl.context.session
    RUNTIME.register_session(thread_id=session.thread_id, session_id=session.id)


@cl.on_chat_end
async def _on_chat_end() -> None:
    session = cl.context.session
    RUNTIME.unregister_session(session.id)


@cl.on_message
async def _on_message(message: cl.Message) -> None:
    global _APP_CLOSED_WARNING_EMITTED, _WORKER_FAILURE_UI_NOTIFIED

    session = cl.context.session
    RUNTIME.register_session(thread_id=session.thread_id, session_id=session.id)

    incoming = IncomingMessage(
        thread_id=session.thread_id,
        session_id=session.id,
        message_id=message.id,
        content=message.content or "",
        author=message.author,
        created_at=message.created_at,
        metadata=message.metadata or {},
    )
    try:
        RUNTIME.enqueue_incoming(incoming)
    except AppClosedError:
        client = RUNTIME.get_client()
        worker_error = client.peek_worker_error() if client is not None else None
        if worker_error is None:
            raise

        summary = _summarize_worker_error(worker_error)
        if not _APP_CLOSED_WARNING_EMITTED:
            LOGGER.warning(
                "Worker app already closed after run_func crash; server shutdown in progress: %s",
                summary,
            )
            _APP_CLOSED_WARNING_EMITTED = True

        if not _WORKER_FAILURE_UI_NOTIFIED:
            try:
                await cl.Message(
                    content=(
                        "Internal worker error detected. Server is shutting down.\n"
                        f"Reason: {summary}"
                    ),
                    author="Easierlit",
                ).send()
                _WORKER_FAILURE_UI_NOTIFIED = True
            except Exception:
                LOGGER.exception("Failed to send worker crash summary message.")



from __future__ import annotations

from typing import Any, Literal

from pydantic import BaseModel, Field


class IncomingMessage(BaseModel):
    thread_id: str
    session_id: str
    message_id: str
    content: str
    author: str
    created_at: str | None = None
    metadata: dict[str, Any] = Field(default_factory=dict)


class OutgoingCommand(BaseModel):
    command: Literal[
        "add_message",
        "add_tool",
        "update_message",
        "update_tool",
        "delete",
        "close",
    ]
    thread_id: str | None = None
    message_id: str | None = None
    content: str | None = None
    author: str = "Assistant"
    metadata: dict[str, Any] = Field(default_factory=dict)



import importlib
import os
import signal

from chainlit.config import config

from easierlit import (
    EasierlitAuthConfig,
    EasierlitClient,
    EasierlitPersistenceConfig,
    EasierlitServer,
)
from easierlit.runtime import get_runtime


def test_serve_forces_headless_and_sidebar(monkeypatch):
    called = {"count": 0, "target": None}
    generated_secret = "s" * 64

    def fake_run_chainlit(target: str):
        called["count"] += 1
        called["target"] = target
        assert config.run.headless is True
        assert config.ui.default_sidebar_state == "open"
        assert config.ui.cot == "full"

        runtime = get_runtime()
        assert runtime.get_auth() is not None
        assert runtime.get_persistence() is not None
        assert runtime.get_persistence().sqlite_path == ".chainlit/test.db"

    chainlit_cli = importlib.import_module("chainlit.cli")
    monkeypatch.setattr(chainlit_cli, "run_chainlit", fake_run_chainlit)
    monkeypatch.setattr("easierlit.server.ensure_jwt_secret", lambda: generated_secret)

    client = EasierlitClient(run_func=lambda _app: None, worker_mode="thread")
    auth = EasierlitAuthConfig(
        username="admin",
        password="admin",
    )
    persistence = EasierlitPersistenceConfig(enabled=True, sqlite_path=".chainlit/test.db")
    server = EasierlitServer(
        client=client,
        host="0.0.0.0",
        port=9000,
        root_path="/custom",
        auth=auth,
        persistence=persistence,
    )
    server.serve()

    assert called["count"] == 1
    assert called["target"].endswith("chainlit_entry.py")
    assert os.environ["CHAINLIT_HOST"] == "0.0.0.0"
    assert os.environ["CHAINLIT_PORT"] == "9000"
    assert os.environ["CHAINLIT_ROOT_PATH"] == "/custom"
    assert os.environ["CHAINLIT_AUTH_COOKIE_NAME"] == "easierlit_access_token"
    assert os.environ["CHAINLIT_AUTH_SECRET"] == generated_secret
    assert len(os.environ["CHAINLIT_AUTH_SECRET"].encode("utf-8")) >= 32


def test_runtime_is_unbound_after_serve(monkeypatch):
    chainlit_cli = importlib.import_module("chainlit.cli")
    monkeypatch.setattr(chainlit_cli, "run_chainlit", lambda _target: None)
    monkeypatch.setattr("easierlit.server.ensure_jwt_secret", lambda: "x" * 64)

    client = EasierlitClient(run_func=lambda _app: None, worker_mode="thread")
    server = EasierlitServer(client=client)
    server.serve()

    runtime = get_runtime()
    assert runtime.get_client() is None
    assert runtime.get_app() is None
    assert runtime.get_auth() is None
    assert runtime.get_persistence() is None


def test_worker_crash_triggers_single_sigint(monkeypatch):
    kill_calls: list[tuple[int, int]] = []

    def fake_kill(pid: int, sig: int):
        kill_calls.append((pid, sig))

    chainlit_cli = importlib.import_module("chainlit.cli")
    monkeypatch.setattr(chainlit_cli, "run_chainlit", lambda _target: None)
    monkeypatch.setattr("easierlit.server.ensure_jwt_secret", lambda: "x" * 64)
    monkeypatch.setattr(os, "kill", fake_kill)

    client = EasierlitClient(run_func=lambda _app: None, worker_mode="thread")
    server = EasierlitServer(client=client)

    def fake_run_chainlit(_target: str):
        handler = client._worker_crash_handler
        assert handler is not None
        handler("Traceback (most recent call last):\nRuntimeError: first")
        handler("Traceback (most recent call last):\nRuntimeError: second")

    monkeypatch.setattr(chainlit_cli, "run_chainlit", fake_run_chainlit)
    server.serve()

    assert len(kill_calls) == 1
    assert kill_calls[0][0] == os.getpid()
    assert kill_calls[0][1] == signal.SIGINT


def test_worker_crash_falls_back_to_sigterm(monkeypatch):
    kill_calls: list[tuple[int, int]] = []

    def fake_kill(pid: int, sig: int):
        kill_calls.append((pid, sig))
        if sig == signal.SIGINT:
            raise OSError("SIGINT failed")

    chainlit_cli = importlib.import_module("chainlit.cli")
    monkeypatch.setattr(chainlit_cli, "run_chainlit", lambda _target: None)
    monkeypatch.setattr("easierlit.server.ensure_jwt_secret", lambda: "x" * 64)
    monkeypatch.setattr(os, "kill", fake_kill)

    client = EasierlitClient(run_func=lambda _app: None, worker_mode="thread")
    server = EasierlitServer(client=client)

    def fake_run_chainlit(_target: str):
        handler = client._worker_crash_handler
        assert handler is not None
        handler("Traceback (most recent call last):\nRuntimeError: fail")

    monkeypatch.setattr(chainlit_cli, "run_chainlit", fake_run_chainlit)
    server.serve()

    assert len(kill_calls) == 2
    assert kill_calls[0][1] == signal.SIGINT
    assert kill_calls[1][1] == signal.SIGTERM



from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass(slots=True)
class EasierlitAuthConfig:
    username: str
    password: str
    identifier: str | None = None
    metadata: dict[str, Any] | None = None

    def __post_init__(self) -> None:
        if not self.username or not self.username.strip():
            raise ValueError("EasierlitAuthConfig.username must not be empty.")
        if not self.password or not self.password.strip():
            raise ValueError("EasierlitAuthConfig.password must not be empty.")


@dataclass(slots=True)
class EasierlitPersistenceConfig:
    enabled: bool = True
    sqlite_path: str = ".chainlit/easierlit.db"



import sqlite3

from easierlit.sqlite_bootstrap import ensure_sqlite_schema


def test_ensure_sqlite_schema_creates_tables_indexes_and_required_columns(tmp_path):
    db_path = tmp_path / "easierlit.db"

    first = ensure_sqlite_schema(db_path)
    second = ensure_sqlite_schema(db_path)

    assert first == second
    assert first.exists()

    connection = sqlite3.connect(first)
    try:
        tables = {
            row[0]
            for row in connection.execute(
                "SELECT name FROM sqlite_master WHERE type='table'"
            ).fetchall()
        }
        assert {"users", "threads", "steps", "elements", "feedbacks"}.issubset(tables)

        step_columns = {
            row[1] for row in connection.execute("PRAGMA table_info('steps')").fetchall()
        }
        assert {"command", "modes", "defaultOpen"}.issubset(step_columns)

        indexes = {
            row[0]
            for row in connection.execute(
                "SELECT name FROM sqlite_master WHERE type='index'"
            ).fetchall()
        }
        assert {
            "idx_steps_thread_id",
            "idx_elements_thread_id",
            "idx_feedbacks_for_id",
            "idx_threads_user_id",
        }.issubset(indexes)
    finally:
        connection.close()


def test_ensure_sqlite_schema_recreates_legacy_database_and_keeps_backup(tmp_path):
    db_path = tmp_path / "easierlit-legacy.db"

    legacy_schema = """
    CREATE TABLE users (
        "id" TEXT PRIMARY KEY,
        "identifier" TEXT NOT NULL UNIQUE,
        "createdAt" TEXT NOT NULL,
        "metadata" TEXT
    );
    CREATE TABLE threads (
        "id" TEXT PRIMARY KEY,
        "createdAt" TEXT,
        "name" TEXT,
        "userId" TEXT,
        "userIdentifier" TEXT,
        "tags" TEXT,
        "metadata" TEXT
    );
    CREATE TABLE steps (
        "id" TEXT PRIMARY KEY,
        "name" TEXT,
        "type" TEXT,
        "threadId" TEXT,
        "parentId" TEXT,
        "streaming" INTEGER,
        "waitForAnswer" INTEGER,
        "isError" INTEGER,
        "metadata" TEXT,
        "tags" TEXT,
        "input" TEXT,
        "output" TEXT,
        "createdAt" TEXT,
        "start" TEXT,
        "end" TEXT,
        "generation" TEXT,
        "showInput" TEXT,
        "language" TEXT
    );
    CREATE TABLE elements (
        "id" TEXT PRIMARY KEY,
        "threadId" TEXT,
        "type" TEXT,
        "chainlitKey" TEXT,
        "url" TEXT,
        "objectKey" TEXT,
        "name" TEXT,
        "display" TEXT,
        "size" INTEGER,
        "language" TEXT,
        "page" INTEGER,
        "autoPlay" INTEGER,
        "playerConfig" TEXT,
        "forId" TEXT,
        "mime" TEXT,
        "props" TEXT
    );
    CREATE TABLE feedbacks (
        "id" TEXT PRIMARY KEY,
        "forId" TEXT,
        "threadId" TEXT,
        "value" REAL,
        "comment" TEXT
    );
    """
    connection = sqlite3.connect(db_path)
    try:
        connection.executescript(legacy_schema)
        connection.commit()
    finally:
        connection.close()

    ensure_sqlite_schema(db_path)
    backups = sorted(tmp_path.glob("easierlit-legacy.db.bak*"))
    assert len(backups) == 1

    backup_connection = sqlite3.connect(backups[0])
    try:
        backup_step_columns = {
            row[1]
            for row in backup_connection.execute("PRAGMA table_info('steps')").fetchall()
        }
        assert "defaultOpen" not in backup_step_columns
    finally:
        backup_connection.close()

    current_connection = sqlite3.connect(db_path)
    try:
        current_step_columns = {
            row[1]
            for row in current_connection.execute("PRAGMA table_info('steps')").fetchall()
        }
        assert {"command", "modes", "defaultOpen"}.issubset(current_step_columns)
    finally:
        current_connection.close()

    ensure_sqlite_schema(db_path)
    backups_after = sorted(tmp_path.glob("easierlit-legacy.db.bak*"))
    assert len(backups_after) == 1



from __future__ import annotations

import os
import secrets
from pathlib import Path

DEFAULT_JWT_SECRET_PATH = ".chainlit/jwt.secret"
DEFAULT_MIN_BYTES = 32


def _generate_secret(min_bytes: int) -> str:
    # token_hex(n) creates a 2n-length ASCII secret.
    token_bytes = max(DEFAULT_MIN_BYTES, (min_bytes + 1) // 2)
    return secrets.token_hex(token_bytes)


def _apply_secure_permissions(path: Path) -> None:
    if os.name == "nt":
        return
    try:
        os.chmod(path, 0o600)
    except OSError:
        # Best effort only: permission changes can fail on some filesystems.
        pass


def ensure_jwt_secret(
    secret_path: str | Path = DEFAULT_JWT_SECRET_PATH,
    min_bytes: int = DEFAULT_MIN_BYTES,
) -> str:
    path = Path(secret_path)
    if not path.is_absolute():
        path = Path.cwd() / path
    path.parent.mkdir(parents=True, exist_ok=True)

    secret: str | None = None
    if path.exists():
        secret = path.read_text(encoding="utf-8").strip()
        if len(secret.encode("utf-8")) < min_bytes:
            secret = None

    if secret is None:
        secret = _generate_secret(min_bytes=min_bytes)
        path.write_text(secret + "\n", encoding="utf-8")

    _apply_secure_permissions(path)
    return secret



import gc
import threading

import pytest

from easierlit import EasierlitApp, EasierlitClient, IncomingMessage, RunFuncExecutionError


def _single_message_worker(app: EasierlitApp) -> None:
    incoming = app.recv(timeout=2.0)
    app.add_message(incoming.thread_id, incoming.content.upper(), author="Worker")


async def _single_message_async_worker(app: EasierlitApp) -> None:
    incoming = await app.arecv(timeout=2.0)
    app.add_message(incoming.thread_id, incoming.content.upper(), author="Worker")


def _crashing_worker(_app: EasierlitApp) -> None:
    raise RuntimeError("worker crashed")


def _sync_noop_worker(_app: EasierlitApp) -> None:
    return None


def test_invalid_worker_mode_raises_value_error():
    with pytest.raises(ValueError, match="worker_mode"):
        EasierlitClient(run_func=_sync_noop_worker, worker_mode="process")  # type: ignore[arg-type]


def test_invalid_run_func_mode_raises_value_error():
    with pytest.raises(ValueError, match="run_func_mode"):
        EasierlitClient(run_func=_sync_noop_worker, run_func_mode="invalid")  # type: ignore[arg-type]


def test_thread_worker_run_and_stop():
    app = EasierlitApp()
    client = EasierlitClient(run_func=_single_message_worker, worker_mode="thread")
    client.run(app)

    app._enqueue_incoming(
        IncomingMessage(
            thread_id="thread-1",
            session_id="session-1",
            message_id="msg-1",
            content="hello",
            author="User",
        )
    )

    command = app._pop_outgoing(timeout=3.0)
    assert command.command == "add_message"
    assert command.content == "HELLO"

    client.stop()


def test_thread_async_worker_auto_mode_run_and_stop():
    app = EasierlitApp()
    client = EasierlitClient(run_func=_single_message_async_worker, worker_mode="thread")
    client.run(app)

    app._enqueue_incoming(
        IncomingMessage(
            thread_id="thread-3",
            session_id="session-3",
            message_id="msg-3",
            content="hey",
            author="User",
        )
    )

    command = app._pop_outgoing(timeout=3.0)
    assert command.command == "add_message"
    assert command.content == "HEY"

    client.stop()


def test_worker_exception_is_raised_on_stop():
    app = EasierlitApp()
    client = EasierlitClient(run_func=_crashing_worker, worker_mode="thread")
    client.run(app)

    try:
        client.stop()
        assert False, "Expected stop() to raise run_func execution error."
    except RunFuncExecutionError:
        pass


def test_run_func_mode_sync_rejects_async_run_func():
    app = EasierlitApp()
    client = EasierlitClient(
        run_func=_single_message_async_worker,
        worker_mode="thread",
        run_func_mode="sync",
    )
    client.run(app)

    with pytest.raises(RunFuncExecutionError, match="run_func_mode='sync'"):
        client.stop()


def test_run_func_mode_async_rejects_sync_run_func():
    app = EasierlitApp()
    client = EasierlitClient(
        run_func=_sync_noop_worker,
        worker_mode="thread",
        run_func_mode="async",
    )
    client.run(app)

    with pytest.raises(RunFuncExecutionError, match="run_func_mode='async'"):
        client.stop()


def test_sync_mode_async_run_func_emits_no_never_awaited_warning(recwarn):
    app = EasierlitApp()
    client = EasierlitClient(
        run_func=_single_message_async_worker,
        worker_mode="thread",
        run_func_mode="sync",
    )
    client.run(app)

    with pytest.raises(RunFuncExecutionError):
        client.stop()

    gc.collect()
    never_awaited = [
        warning
        for warning in recwarn.list
        if "was never awaited" in str(warning.message)
    ]
    assert not never_awaited


def test_thread_worker_records_error_and_invokes_crash_handler():
    app = EasierlitApp()
    crash_event = threading.Event()
    crash_payloads: list[str] = []

    client = EasierlitClient(run_func=_crashing_worker, worker_mode="thread")
    client.set_worker_crash_handler(
        lambda traceback_text: (
            crash_payloads.append(traceback_text),
            crash_event.set(),
        )
    )
    client.run(app)

    assert crash_event.wait(timeout=2.0)
    error = client.peek_worker_error()
    assert error is not None
    assert "worker crashed" in error
    assert len(crash_payloads) == 1

    with pytest.raises(RunFuncExecutionError):
        client.stop()



from __future__ import annotations

import logging
import sqlite3
from pathlib import Path

LOGGER = logging.getLogger(__name__)


SCHEMA_SQL = """
CREATE TABLE IF NOT EXISTS users (
    "id" TEXT PRIMARY KEY,
    "identifier" TEXT NOT NULL UNIQUE,
    "createdAt" TEXT NOT NULL,
    "metadata" TEXT
);

CREATE TABLE IF NOT EXISTS threads (
    "id" TEXT PRIMARY KEY,
    "createdAt" TEXT,
    "name" TEXT,
    "userId" TEXT,
    "userIdentifier" TEXT,
    "tags" TEXT,
    "metadata" TEXT
);

CREATE TABLE IF NOT EXISTS steps (
    "id" TEXT PRIMARY KEY,
    "name" TEXT,
    "type" TEXT,
    "threadId" TEXT,
    "parentId" TEXT,
    "command" TEXT,
    "modes" TEXT,
    "streaming" INTEGER,
    "waitForAnswer" INTEGER,
    "isError" INTEGER,
    "metadata" TEXT,
    "tags" TEXT,
    "input" TEXT,
    "output" TEXT,
    "createdAt" TEXT,
    "start" TEXT,
    "end" TEXT,
    "generation" TEXT,
    "defaultOpen" INTEGER,
    "showInput" TEXT,
    "language" TEXT
);

CREATE TABLE IF NOT EXISTS elements (
    "id" TEXT PRIMARY KEY,
    "threadId" TEXT,
    "type" TEXT,
    "chainlitKey" TEXT,
    "url" TEXT,
    "objectKey" TEXT,
    "name" TEXT,
    "display" TEXT,
    "size" INTEGER,
    "language" TEXT,
    "page" INTEGER,
    "autoPlay" INTEGER,
    "playerConfig" TEXT,
    "forId" TEXT,
    "mime" TEXT,
    "props" TEXT
);

CREATE TABLE IF NOT EXISTS feedbacks (
    "id" TEXT PRIMARY KEY,
    "forId" TEXT,
    "threadId" TEXT,
    "value" REAL,
    "comment" TEXT
);

CREATE INDEX IF NOT EXISTS idx_steps_thread_id ON steps("threadId");
CREATE INDEX IF NOT EXISTS idx_elements_thread_id ON elements("threadId");
CREATE INDEX IF NOT EXISTS idx_feedbacks_for_id ON feedbacks("forId");
CREATE INDEX IF NOT EXISTS idx_threads_user_id ON threads("userId");
"""

REQUIRED_COLUMNS: dict[str, set[str]] = {
    "users": {"id", "identifier", "createdAt", "metadata"},
    "threads": {"id", "createdAt", "name", "userId", "userIdentifier", "tags", "metadata"},
    "steps": {
        "id",
        "name",
        "type",
        "threadId",
        "parentId",
        "command",
        "modes",
        "streaming",
        "waitForAnswer",
        "isError",
        "metadata",
        "tags",
        "input",
        "output",
        "createdAt",
        "start",
        "end",
        "generation",
        "defaultOpen",
        "showInput",
        "language",
    },
    "elements": {
        "id",
        "threadId",
        "type",
        "chainlitKey",
        "url",
        "objectKey",
        "name",
        "display",
        "size",
        "language",
        "page",
        "autoPlay",
        "playerConfig",
        "forId",
        "mime",
        "props",
    },
    "feedbacks": {"id", "forId", "threadId", "value", "comment"},
}


def _table_columns(connection: sqlite3.Connection, table_name: str) -> set[str]:
    rows = connection.execute(f'PRAGMA table_info("{table_name}")').fetchall()
    return {str(row[1]) for row in rows}


def _is_schema_compatible(connection: sqlite3.Connection) -> bool:
    for table_name, required_columns in REQUIRED_COLUMNS.items():
        columns = _table_columns(connection, table_name)
        if not columns:
            return False
        if not required_columns.issubset(columns):
            return False
    return True


def _next_backup_path(path: Path) -> Path:
    candidate = path.with_name(f"{path.name}.bak")
    if not candidate.exists():
        return candidate

    index = 1
    while True:
        candidate = path.with_name(f"{path.name}.bak{index}")
        if not candidate.exists():
            return candidate
        index += 1


def _recreate_if_incompatible(path: Path) -> None:
    if not path.exists():
        return

    needs_recreate = False
    try:
        connection = sqlite3.connect(path)
        try:
            needs_recreate = not _is_schema_compatible(connection)
        finally:
            connection.close()
    except sqlite3.DatabaseError:
        needs_recreate = True

    if not needs_recreate:
        return

    backup_path = _next_backup_path(path)
    path.replace(backup_path)
    LOGGER.warning(
        "Detected incompatible SQLite schema at %s. Recreated database; backup saved to %s.",
        path,
        backup_path,
    )


def ensure_sqlite_schema(sqlite_path: str | Path) -> Path:
    path = Path(sqlite_path)
    if not path.is_absolute():
        path = Path.cwd() / path

    path.parent.mkdir(parents=True, exist_ok=True)
    _recreate_if_incompatible(path)

    connection = sqlite3.connect(path)
    try:
        connection.executescript(SCHEMA_SQL)
        connection.commit()
    finally:
        connection.close()

    return path



from __future__ import annotations

import logging
import os
import signal
import threading
from pathlib import Path

from .app import EasierlitApp
from .client import EasierlitClient
from .jwt_secret import ensure_jwt_secret
from .runtime import get_runtime
from .settings import EasierlitAuthConfig, EasierlitPersistenceConfig

LOGGER = logging.getLogger(__name__)


class EasierlitServer:
    def __init__(
        self,
        client: EasierlitClient,
        host: str = "127.0.0.1",
        port: int = 8000,
        root_path: str = "",
        auth: EasierlitAuthConfig | None = None,
        persistence: EasierlitPersistenceConfig | None = None,
    ):
        self.client = client
        self.host = host
        self.port = port
        self.root_path = root_path
        self.auth = auth
        self.persistence = persistence or EasierlitPersistenceConfig()

    def serve(self) -> None:
        from chainlit.cli import run_chainlit
        from chainlit.config import config

        app = EasierlitApp()
        runtime = get_runtime()
        shutdown_requested = threading.Event()

        runtime.bind(
            client=self.client,
            app=app,
            auth=self.auth,
            persistence=self.persistence,
        )

        def _handle_worker_crash(traceback_text: str) -> None:
            summary = "Unknown run_func error"
            lines = traceback_text.strip().splitlines()
            if lines:
                summary = lines[-1]

            LOGGER.error("run_func crashed: %s", summary)
            LOGGER.error("run_func traceback:\n%s", traceback_text)

            if shutdown_requested.is_set():
                return
            shutdown_requested.set()

            pid = os.getpid()
            try:
                os.kill(pid, signal.SIGINT)
            except Exception:
                LOGGER.exception("Failed to send SIGINT to stop server; retrying with SIGTERM.")
                try:
                    os.kill(pid, signal.SIGTERM)
                except Exception:
                    LOGGER.exception("Failed to send SIGTERM after SIGINT failure.")

        self.client.set_worker_crash_handler(_handle_worker_crash)
        self.client.run(app)

        os.environ["CHAINLIT_HOST"] = self.host
        os.environ["CHAINLIT_PORT"] = str(self.port)
        os.environ["CHAINLIT_ROOT_PATH"] = self.root_path
        os.environ["CHAINLIT_AUTH_COOKIE_NAME"] = "easierlit_access_token"
        os.environ["CHAINLIT_AUTH_SECRET"] = ensure_jwt_secret()

        # Easierlit policy: always run headless, always keep sidebar open, and show full CoT.
        config.run.headless = True
        config.ui.default_sidebar_state = "open"
        config.ui.cot = "full"

        entrypoint = self._entrypoint_path()

        try:
            run_chainlit(str(entrypoint))
        finally:
            self.client.set_worker_crash_handler(None)
            self.client.stop()
            runtime.unbind()

    def _entrypoint_path(self) -> Path:
        return Path(__file__).resolve().parent / "chainlit_entry.py"



import asyncio

import pytest

from easierlit import AppClosedError, EasierlitApp, IncomingMessage


def test_recv_add_update_delete_flow():
    app = EasierlitApp()

    incoming = IncomingMessage(
        thread_id="thread-1",
        session_id="session-1",
        message_id="msg-1",
        content="hello",
        author="User",
    )
    app._enqueue_incoming(incoming)

    received = app.recv(timeout=1.0)
    assert received.thread_id == "thread-1"
    assert received.content == "hello"

    message_id = app.add_message(thread_id="thread-1", content="world", author="Bot")
    add_cmd = app._pop_outgoing(timeout=1.0)
    assert add_cmd.command == "add_message"
    assert add_cmd.message_id == message_id
    assert add_cmd.content == "world"

    app.update_message(thread_id="thread-1", message_id=message_id, content="new")
    update_cmd = app._pop_outgoing(timeout=1.0)
    assert update_cmd.command == "update_message"
    assert update_cmd.message_id == message_id
    assert update_cmd.content == "new"

    app.delete_message(thread_id="thread-1", message_id=message_id)
    delete_cmd = app._pop_outgoing(timeout=1.0)
    assert delete_cmd.command == "delete"
    assert delete_cmd.message_id == message_id


def test_tool_and_thought_enqueue_flow():
    app = EasierlitApp()

    tool_message_id = app.add_tool(
        thread_id="thread-1",
        tool_name="SearchTool",
        content='{"query":"chainlit"}',
    )
    tool_add_cmd = app._pop_outgoing(timeout=1.0)
    assert tool_add_cmd.command == "add_tool"
    assert tool_add_cmd.message_id == tool_message_id
    assert tool_add_cmd.author == "SearchTool"
    assert tool_add_cmd.content == '{"query":"chainlit"}'

    thought_message_id = app.add_thought(
        thread_id="thread-1",
        content="I should call a retrieval tool first.",
    )
    thought_add_cmd = app._pop_outgoing(timeout=1.0)
    assert thought_add_cmd.command == "add_tool"
    assert thought_add_cmd.message_id == thought_message_id
    assert thought_add_cmd.author == "Reasoning"
    assert thought_add_cmd.content == "I should call a retrieval tool first."

    app.update_tool(
        thread_id="thread-1",
        message_id=tool_message_id,
        tool_name="SearchTool",
        content='{"results":3}',
    )
    tool_update_cmd = app._pop_outgoing(timeout=1.0)
    assert tool_update_cmd.command == "update_tool"
    assert tool_update_cmd.message_id == tool_message_id
    assert tool_update_cmd.author == "SearchTool"
    assert tool_update_cmd.content == '{"results":3}'

    app.update_thought(
        thread_id="thread-1",
        message_id=thought_message_id,
        content="Now I can synthesize the final answer.",
    )
    thought_update_cmd = app._pop_outgoing(timeout=1.0)
    assert thought_update_cmd.command == "update_tool"
    assert thought_update_cmd.message_id == thought_message_id
    assert thought_update_cmd.author == "Reasoning"
    assert thought_update_cmd.content == "Now I can synthesize the final answer."


def test_recv_timeout_and_close():
    app = EasierlitApp()

    try:
        app.recv(timeout=0.05)
        assert False, "recv() should timeout when no message exists."
    except TimeoutError:
        pass

    app.close()

    try:
        app.recv(timeout=0.05)
        assert False, "recv() should fail once app is closed."
    except AppClosedError:
        pass


def test_arecv_flow():
    app = EasierlitApp()
    incoming = IncomingMessage(
        thread_id="thread-2",
        session_id="session-2",
        message_id="msg-2",
        content="hello async",
        author="User",
    )
    app._enqueue_incoming(incoming)

    received = asyncio.run(app.arecv(timeout=1.0))
    assert received.thread_id == "thread-2"
    assert received.content == "hello async"


def test_arecv_timeout_and_close():
    app = EasierlitApp()

    with pytest.raises(TimeoutError):
        asyncio.run(app.arecv(timeout=0.05))

    app.close()

    with pytest.raises(AppClosedError):
        asyncio.run(app.arecv(timeout=0.05))



from __future__ import annotations

import asyncio
import inspect
import queue
import threading
import traceback
from typing import Any, Callable, Literal

from .app import EasierlitApp
from .errors import RunFuncExecutionError, WorkerAlreadyRunningError

RunFuncMode = Literal["auto", "sync", "async"]


def _close_unawaited_awaitable(value: Any) -> None:
    close = getattr(value, "close", None)
    if callable(close):
        try:
            close()
        except Exception:
            pass

    cancel = getattr(value, "cancel", None)
    if callable(cancel):
        try:
            cancel()
        except Exception:
            pass


def _run_awaitable(awaitable: Any) -> None:
    if inspect.iscoroutine(awaitable):
        asyncio.run(awaitable)
        return

    async def _await_result() -> None:
        await awaitable

    asyncio.run(_await_result())


def _execute_run_func(
    run_func: Callable[[EasierlitApp], Any],
    app: EasierlitApp,
    run_func_mode: RunFuncMode,
) -> None:
    result = run_func(app)
    is_awaitable = inspect.isawaitable(result)

    if run_func_mode == "sync":
        if is_awaitable:
            _close_unawaited_awaitable(result)
            raise TypeError(
                "run_func_mode='sync' requires a synchronous run_func "
                "that does not return an awaitable."
            )
        return

    if run_func_mode == "async":
        if not is_awaitable:
            raise TypeError(
                "run_func_mode='async' requires run_func to return an awaitable."
            )
        _run_awaitable(result)
        return

    if is_awaitable:
        _run_awaitable(result)


class EasierlitClient:
    def __init__(
        self,
        run_func: Callable[[EasierlitApp], Any],
        worker_mode: Literal["thread"] = "thread",
        run_func_mode: RunFuncMode = "auto",
    ):
        if worker_mode != "thread":
            raise ValueError("worker_mode must be 'thread'.")
        if run_func_mode not in ("auto", "sync", "async"):
            raise ValueError("run_func_mode must be one of 'auto', 'sync', or 'async'.")

        self.run_func = run_func
        self.worker_mode = worker_mode
        self.run_func_mode = run_func_mode

        self._app: EasierlitApp | None = None

        self._thread: threading.Thread | None = None
        self._thread_error_queue: queue.Queue[str] = queue.Queue()

        self._worker_error_traceback: str | None = None
        self._worker_error_lock = threading.Lock()
        self._worker_crash_handler: Callable[[str], None] | None = None

    def run(self, app: EasierlitApp) -> None:
        if self._is_worker_running():
            raise WorkerAlreadyRunningError("run() called while worker is already running.")

        self._reset_worker_error_state()
        self._app = app
        self._start_thread_worker(app)

    def stop(self, timeout: float = 5.0) -> None:
        app = self._app
        if app is not None:
            app.close()

        if self._thread is not None:
            self._thread.join(timeout=timeout)
            self._thread = None

        self._app = None
        self._raise_worker_error_if_any()

    def _start_thread_worker(self, app: EasierlitApp) -> None:
        self._thread = threading.Thread(
            target=self._thread_worker_entry,
            args=(app,),
            daemon=True,
        )
        self._thread.start()

    def _thread_worker_entry(self, app: EasierlitApp) -> None:
        try:
            _execute_run_func(self.run_func, app, self.run_func_mode)
        except Exception:
            traceback_text = traceback.format_exc()
            self._thread_error_queue.put(traceback_text)
            self._record_worker_error(traceback_text)
        finally:
            app.close()

    def _is_worker_running(self) -> bool:
        return self._thread is not None and self._thread.is_alive()

    def _raise_worker_error_if_any(self) -> None:
        worker_error = self.peek_worker_error()
        if worker_error:
            raise RunFuncExecutionError(worker_error)

        if not self._thread_error_queue.empty():
            raise RunFuncExecutionError(self._thread_error_queue.get())

    def set_worker_crash_handler(self, handler: Callable[[str], None] | None) -> None:
        with self._worker_error_lock:
            self._worker_crash_handler = handler

    def peek_worker_error(self) -> str | None:
        with self._worker_error_lock:
            return self._worker_error_traceback

    def _record_worker_error(self, traceback_text: str) -> None:
        with self._worker_error_lock:
            self._worker_error_traceback = traceback_text
            handler = self._worker_crash_handler

        if handler is not None:
            handler(traceback_text)

    def _reset_worker_error_state(self) -> None:
        with self._worker_error_lock:
            self._worker_error_traceback = None

        while True:
            try:
                self._thread_error_queue.get_nowait()
            except queue.Empty:
                break



from __future__ import annotations

import asyncio
import json
import queue
import threading
from typing import Any
from uuid import uuid4

from chainlit.data import get_data_layer
from chainlit.types import Pagination, ThreadFilter
from chainlit.user import User

from .errors import AppClosedError, DataPersistenceNotEnabledError
from .models import IncomingMessage, OutgoingCommand
from .runtime import get_runtime


class EasierlitApp:
    """
    Communication bridge between Chainlit callbacks and user run_func.

    Easierlit runs with a single thread worker. This object is the primary API
    surface for worker message and thread CRUD operations.
    """

    _THOUGHT_TOOL_NAME = "Reasoning"

    def __init__(self):
        self._incoming_queue: queue.Queue[IncomingMessage | None] = queue.Queue()
        self._outgoing_queue: queue.Queue[OutgoingCommand] = queue.Queue()
        self._closed = threading.Event()
        self._runtime = get_runtime()

    def recv(self, timeout: float | None = None) -> IncomingMessage:
        if self._closed.is_set():
            raise AppClosedError("EasierlitApp is closed.")

        try:
            if timeout is None:
                item = self._incoming_queue.get()
            else:
                item = self._incoming_queue.get(timeout=timeout)
        except queue.Empty as exc:
            raise TimeoutError("Timed out waiting for an incoming message.") from exc

        if item is None:
            raise AppClosedError("EasierlitApp is closed.")

        return item

    async def arecv(self, timeout: float | None = None) -> IncomingMessage:
        return await asyncio.to_thread(self.recv, timeout)

    def add_message(
        self,
        thread_id: str,
        content: str,
        author: str = "Assistant",
        metadata: dict | None = None,
    ) -> str:
        """Enqueue an assistant message and return the generated message id."""
        message_id = str(uuid4())
        self._put_outgoing(
            OutgoingCommand(
                command="add_message",
                thread_id=thread_id,
                message_id=message_id,
                content=content,
                author=author,
                metadata=metadata or {},
            )
        )
        return message_id

    def add_tool(
        self,
        thread_id: str,
        tool_name: str,
        content: str,
        metadata: dict | None = None,
    ) -> str:
        """Enqueue a tool-call step and return the generated message id."""
        message_id = str(uuid4())
        self._put_outgoing(
            OutgoingCommand(
                command="add_tool",
                thread_id=thread_id,
                message_id=message_id,
                content=content,
                author=tool_name,
                metadata=metadata or {},
            )
        )
        return message_id

    def add_thought(
        self,
        thread_id: str,
        content: str,
        metadata: dict | None = None,
    ) -> str:
        """Enqueue a reasoning step as a tool-call step."""
        return self.add_tool(
            thread_id=thread_id,
            tool_name=self._THOUGHT_TOOL_NAME,
            content=content,
            metadata=metadata,
        )

    def update_message(
        self,
        thread_id: str,
        message_id: str,
        content: str,
        metadata: dict | None = None,
    ) -> None:
        self._put_outgoing(
            OutgoingCommand(
                command="update_message",
                thread_id=thread_id,
                message_id=message_id,
                content=content,
                metadata=metadata or {},
            )
        )

    def update_tool(
        self,
        thread_id: str,
        message_id: str,
        tool_name: str,
        content: str,
        metadata: dict | None = None,
    ) -> None:
        self._put_outgoing(
            OutgoingCommand(
                command="update_tool",
                thread_id=thread_id,
                message_id=message_id,
                content=content,
                author=tool_name,
                metadata=metadata or {},
            )
        )

    def update_thought(
        self,
        thread_id: str,
        message_id: str,
        content: str,
        metadata: dict | None = None,
    ) -> None:
        self.update_tool(
            thread_id=thread_id,
            message_id=message_id,
            tool_name=self._THOUGHT_TOOL_NAME,
            content=content,
            metadata=metadata,
        )

    def delete_message(self, thread_id: str, message_id: str) -> None:
        self._put_outgoing(
            OutgoingCommand(
                command="delete",
                thread_id=thread_id,
                message_id=message_id,
            )
        )

    def list_threads(
        self,
        first: int = 20,
        cursor: str | None = None,
        search: str | None = None,
        user_identifier: str | None = None,
    ):
        data_layer = self._get_data_layer_or_raise()

        async def _list_threads():
            user_id = None
            if user_identifier is not None:
                persisted_user = await data_layer.get_user(user_identifier)
                if persisted_user is None:
                    raise ValueError(f"User '{user_identifier}' not found.")
                user_id = persisted_user.id

            pagination = Pagination(first=first, cursor=cursor)
            filters = ThreadFilter(search=search, userId=user_id)
            threads = await data_layer.list_threads(pagination, filters)
            return self._normalize_threads_tags(threads)

        return self._runtime.run_coroutine_sync(_list_threads())

    def get_thread(self, thread_id: str) -> dict:
        data_layer = self._get_data_layer_or_raise()

        async def _get_thread():
            thread = await data_layer.get_thread(thread_id)
            if thread is None:
                raise ValueError(f"Thread '{thread_id}' not found.")
            return self._normalize_thread_tags(thread)

        return self._runtime.run_coroutine_sync(_get_thread())

    def get_history(self, thread_id: str) -> dict:
        """
        Return thread metadata and one ordered history list.

        `items` preserves the original order in `thread["steps"]` and includes
        both message steps and non-message steps.
        """
        thread = self.get_thread(thread_id)

        raw_steps = thread.get("steps")
        step_items = raw_steps if isinstance(raw_steps, list) else []
        items = [item for item in step_items if isinstance(item, dict)]

        thread_metadata = dict(thread)
        thread_metadata.pop("steps", None)

        return {
            "thread": thread_metadata,
            "items": items,
        }

    def timeline(self, thread_id: str) -> dict:
        """Backward-compatible alias for `get_history`."""
        return self.get_history(thread_id)

    def get_thread_timeline(self, thread_id: str) -> dict:
        """Backward-compatible alias for `get_history`."""
        return self.get_history(thread_id)

    def get_thread_messages_and_steps(self, thread_id: str) -> dict:
        """Backward-compatible alias for `get_history`."""
        return self.get_history(thread_id)

    def get_timeline(self, thread_id: str) -> dict:
        """Backward-compatible alias for `get_history`."""
        return self.get_history(thread_id)

    def update_thread(
        self,
        thread_id: str,
        name: str | None = None,
        metadata: dict | None = None,
        tags: list[str] | None = None,
    ) -> None:
        self._write_thread(
            thread_id=thread_id,
            name=name,
            metadata=metadata,
            tags=tags,
            require_existing=True,
        )

    def new_thread(
        self,
        name: str | None = None,
        metadata: dict | None = None,
        tags: list[str] | None = None,
    ) -> str:
        data_layer = self._get_data_layer_or_raise()
        prepared_tags = self._prepare_tags_for_update(tags, data_layer)

        async def _new_thread() -> str:
            owner_user_id = await self._resolve_default_owner_user_id(data_layer)
            for _ in range(16):
                thread_id = str(uuid4())
                existing = await data_layer.get_thread(thread_id)
                if existing is not None:
                    continue

                await data_layer.update_thread(
                    thread_id=thread_id,
                    name=name,
                    user_id=owner_user_id,
                    metadata=metadata,
                    tags=prepared_tags,
                )
                return thread_id

            raise RuntimeError("Failed to allocate unique thread_id.")

        return self._runtime.run_coroutine_sync(_new_thread())

    def _write_thread(
        self,
        thread_id: str,
        name: str | None,
        metadata: dict | None,
        tags: list[str] | None,
        require_existing: bool,
    ) -> None:
        data_layer = self._get_data_layer_or_raise()
        prepared_tags = self._prepare_tags_for_update(tags, data_layer)

        async def _write_thread():
            existing = await data_layer.get_thread(thread_id)
            if require_existing and existing is None:
                raise ValueError(f"Thread '{thread_id}' not found.")
            if not require_existing and existing is not None:
                raise ValueError(f"Thread '{thread_id}' already exists.")

            owner_user_id = await self._resolve_default_owner_user_id(data_layer)
            await data_layer.update_thread(
                thread_id=thread_id,
                name=name,
                user_id=owner_user_id,
                metadata=metadata,
                tags=prepared_tags,
            )

        self._runtime.run_coroutine_sync(_write_thread())

    def delete_thread(self, thread_id: str) -> None:
        data_layer = self._get_data_layer_or_raise()

        async def _delete_thread():
            await data_layer.delete_thread(thread_id)

        self._runtime.run_coroutine_sync(_delete_thread())

    def close(self) -> None:
        if self._closed.is_set():
            return

        self._closed.set()
        self._incoming_queue.put_nowait(None)
        self._outgoing_queue.put_nowait(OutgoingCommand(command="close"))

    def is_closed(self) -> bool:
        return self._closed.is_set()

    def _enqueue_incoming(self, message: IncomingMessage) -> None:
        if self._closed.is_set():
            raise AppClosedError("Cannot enqueue incoming message to a closed app.")
        self._incoming_queue.put_nowait(message)

    def _pop_outgoing(self, timeout: float | None = 0.1) -> OutgoingCommand:
        if timeout is None:
            return self._outgoing_queue.get()
        return self._outgoing_queue.get(timeout=timeout)

    def _put_outgoing(self, command: OutgoingCommand) -> None:
        if self._closed.is_set():
            raise AppClosedError("Cannot send command to a closed app.")
        self._outgoing_queue.put_nowait(command)

    def _get_data_layer_or_raise(self):
        data_layer = get_data_layer()
        if data_layer is None:
            raise DataPersistenceNotEnabledError(
                "Data persistence is not enabled. Configure Chainlit data layer first."
            )
        return data_layer

    def _is_sqlite_sqlalchemy_data_layer(self, data_layer: Any) -> bool:
        conninfo = getattr(data_layer, "_conninfo", None)
        if isinstance(conninfo, str) and conninfo.lower().startswith("sqlite"):
            return True

        engine = getattr(data_layer, "engine", None)
        url = getattr(engine, "url", None)
        drivername = getattr(url, "drivername", None)
        return isinstance(drivername, str) and drivername.lower().startswith("sqlite")

    def _prepare_tags_for_update(
        self,
        tags: list[str] | None,
        data_layer: Any,
    ) -> list[str] | str | None:
        if tags is None:
            return None
        if self._is_sqlite_sqlalchemy_data_layer(data_layer):
            return json.dumps(tags)
        return tags

    async def _resolve_default_owner_user_id(self, data_layer: Any) -> str | None:
        auth = self._runtime.get_auth()
        if auth is None:
            return None

        identifier = auth.identifier or auth.username
        persisted_user = await data_layer.get_user(identifier)
        if persisted_user is not None:
            return persisted_user.id

        if not hasattr(data_layer, "create_user"):
            return None

        created_user = await data_layer.create_user(
            User(
                identifier=identifier,
                metadata=auth.metadata or {},
            )
        )
        if created_user is None:
            return None

        return created_user.id

    def _normalize_thread_tags(self, thread: dict) -> dict:
        tags = thread.get("tags")
        if not isinstance(tags, str):
            return thread

        try:
            decoded = json.loads(tags)
        except json.JSONDecodeError:
            return thread

        if not isinstance(decoded, list):
            return thread

        normalized = dict(thread)
        normalized["tags"] = decoded
        return normalized

    def _normalize_threads_tags(self, threads):
        if not hasattr(threads, "data"):
            return threads

        threads.data = [
            self._normalize_thread_tags(thread)
            if isinstance(thread, dict)
            else thread
            for thread in threads.data
        ]
        return threads



from __future__ import annotations

import asyncio
import logging
import queue
import threading
from typing import TYPE_CHECKING, Any, Coroutine

from chainlit.context import init_http_context
from chainlit.data import get_data_layer
from chainlit.utils import utc_now

from .errors import ThreadSessionNotActiveError
from .models import IncomingMessage, OutgoingCommand

if TYPE_CHECKING:
    from chainlit.session import WebsocketSession

    from .app import EasierlitApp
    from .client import EasierlitClient
    from .settings import EasierlitAuthConfig, EasierlitPersistenceConfig

LOGGER = logging.getLogger(__name__)


class RuntimeRegistry:
    def __init__(self) -> None:
        self._client: EasierlitClient | None = None
        self._app: EasierlitApp | None = None
        self._auth: EasierlitAuthConfig | None = None
        self._persistence: EasierlitPersistenceConfig | None = None

        self._thread_to_session: dict[str, str] = {}
        self._session_to_thread: dict[str, str] = {}

        self._main_loop: asyncio.AbstractEventLoop | None = None
        self._dispatcher_task: asyncio.Task[None] | None = None

        self._lock = threading.RLock()

    def bind(
        self,
        *,
        client: EasierlitClient,
        app: EasierlitApp,
        auth: EasierlitAuthConfig | None = None,
        persistence: EasierlitPersistenceConfig | None = None,
    ) -> None:
        with self._lock:
            self._client = client
            self._app = app
            self._auth = auth
            self._persistence = persistence
            self._thread_to_session.clear()
            self._session_to_thread.clear()

    def unbind(self) -> None:
        with self._lock:
            self._client = None
            self._app = None
            self._auth = None
            self._persistence = None
            self._thread_to_session.clear()
            self._session_to_thread.clear()
            self._main_loop = None
            self._dispatcher_task = None

    def get_client(self) -> EasierlitClient | None:
        return self._client

    def get_app(self) -> EasierlitApp | None:
        return self._app

    def get_auth(self) -> EasierlitAuthConfig | None:
        return self._auth

    def get_persistence(self) -> EasierlitPersistenceConfig | None:
        return self._persistence

    def set_main_loop(self, loop: asyncio.AbstractEventLoop) -> None:
        self._main_loop = loop

    def get_main_loop(self) -> asyncio.AbstractEventLoop | None:
        return self._main_loop

    def run_coroutine_sync(self, coro: Coroutine[Any, Any, Any]) -> Any:
        loop = self._main_loop
        if loop and loop.is_running():
            try:
                running_loop = asyncio.get_running_loop()
            except RuntimeError:
                running_loop = None

            if running_loop is loop:
                raise RuntimeError(
                    "Cannot synchronously wait for a coroutine from the Chainlit event loop."
                )

            future = asyncio.run_coroutine_threadsafe(coro, loop)
            return future.result()

        return asyncio.run(coro)

    def register_session(self, thread_id: str, session_id: str) -> None:
        with self._lock:
            self._thread_to_session[thread_id] = session_id
            self._session_to_thread[session_id] = thread_id

    def unregister_session(self, session_id: str) -> None:
        with self._lock:
            thread_id = self._session_to_thread.pop(session_id, None)
            if thread_id is not None:
                self._thread_to_session.pop(thread_id, None)

    def get_session_id_for_thread(self, thread_id: str) -> str | None:
        with self._lock:
            return self._thread_to_session.get(thread_id)

    def enqueue_incoming(self, message: IncomingMessage) -> None:
        app = self._app
        if app is None:
            return
        app._enqueue_incoming(message)

    async def start_dispatcher(self) -> None:
        if self._app is None:
            raise RuntimeError("No active app bound to runtime.")

        if self._dispatcher_task and not self._dispatcher_task.done():
            return

        self._dispatcher_task = asyncio.create_task(self._dispatch_outgoing_loop())

    async def stop_dispatcher(self) -> None:
        task = self._dispatcher_task
        if task is None:
            return

        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass
        finally:
            self._dispatcher_task = None

    async def _dispatch_outgoing_loop(self) -> None:
        app = self._app
        if app is None:
            return

        while not app.is_closed():
            try:
                command = await asyncio.to_thread(app._pop_outgoing, 0.2)
            except queue.Empty:
                continue

            if command.command == "close":
                break

            try:
                await self.apply_outgoing_command(command)
            except Exception:
                LOGGER.exception(
                    "Failed to apply outgoing command: %s",
                    command.model_dump(),
                )

    def _resolve_session(self, thread_id: str) -> WebsocketSession | None:
        from chainlit.session import WebsocketSession

        session_id = self.get_session_id_for_thread(thread_id)
        if session_id is None:
            return None

        session = WebsocketSession.get_by_id(session_id)
        if session is None:
            self.unregister_session(session_id)
            return None

        return session

    async def apply_outgoing_command(self, command: OutgoingCommand) -> None:
        if command.command == "close":
            return

        thread_id = command.thread_id
        if not thread_id:
            raise ValueError("Outgoing command is missing thread_id.")

        session = self._resolve_session(thread_id)
        if session is not None:
            await self._apply_realtime_command(session, command)
            return

        data_layer = get_data_layer()
        if not data_layer:
            raise ThreadSessionNotActiveError(
                f"Thread '{thread_id}' has no active session and no data layer fallback."
            )

        await self._apply_data_layer_command(command)

    async def _apply_realtime_command(
        self, session: WebsocketSession, command: OutgoingCommand
    ) -> None:
        from chainlit.context import init_ws_context
        from chainlit.message import Message
        from chainlit.step import Step

        init_ws_context(session)

        if command.command == "add_message":
            message = Message(
                id=command.message_id,
                content=command.content or "",
                author=command.author,
                metadata=command.metadata,
            )
            await message.send()
            return

        if command.command == "update_message":
            if not command.message_id:
                raise ValueError("Update command requires message_id.")
            message = Message(
                id=command.message_id,
                content=command.content or "",
                author=command.author,
                metadata=command.metadata,
            )
            await message.update()
            return

        if command.command == "add_tool":
            if not command.message_id:
                raise ValueError("Add tool command requires message_id.")
            step = Step(
                id=command.message_id,
                thread_id=command.thread_id,
                name=command.author,
                type="tool",
                metadata=command.metadata,
            )
            step.output = command.content or ""
            await step.send()
            return

        if command.command == "update_tool":
            if not command.message_id:
                raise ValueError("Update tool command requires message_id.")
            step = Step(
                id=command.message_id,
                thread_id=command.thread_id,
                name=command.author,
                type="tool",
                metadata=command.metadata,
            )
            step.output = command.content or ""
            await step.update()
            return

        if command.command == "delete":
            if not command.message_id:
                raise ValueError("Delete command requires message_id.")
            step = Step(
                id=command.message_id,
                thread_id=command.thread_id,
                name=command.author,
                type="undefined",
                metadata=command.metadata,
            )
            await step.remove()
            return

        raise ValueError(f"Unsupported command: {command.command}")

    async def _apply_data_layer_command(self, command: OutgoingCommand) -> None:
        data_layer = get_data_layer()
        if not data_layer:
            raise RuntimeError("Data layer unexpectedly missing.")

        if not command.thread_id:
            raise ValueError(f"{command.command} command requires thread_id.")
        self._init_data_layer_http_context(command.thread_id)

        if command.command == "delete":
            if not command.message_id:
                raise ValueError("Delete command requires message_id.")
            await data_layer.delete_step(command.message_id)
            return

        if not command.message_id:
            raise ValueError(f"{command.command} command requires message_id.")

        timestamp = utc_now()
        is_tool_command = command.command in ("add_tool", "update_tool")
        step_dict = {
            "id": command.message_id,
            "threadId": command.thread_id,
            "name": command.author,
            "type": "tool" if is_tool_command else "assistant_message",
            "output": command.content or "",
            "createdAt": timestamp,
            "start": timestamp,
            "end": timestamp,
            "streaming": False,
            "isError": False,
            "waitForAnswer": False,
            "metadata": command.metadata,
        }

        if command.command in ("add_message", "add_tool"):
            await data_layer.create_step(step_dict)
            return

        if command.command in ("update_message", "update_tool"):
            await data_layer.update_step(step_dict)
            return

        raise ValueError(f"Unsupported command: {command.command}")

    def _init_data_layer_http_context(self, thread_id: str) -> None:
        init_http_context(thread_id=thread_id, client_type="webapp")


_RUNTIME = RuntimeRegistry()


def get_runtime() -> RuntimeRegistry:
    return _RUNTIME



import asyncio
from types import SimpleNamespace

import pytest

import easierlit.chainlit_entry as chainlit_entry
from easierlit import EasierlitApp, EasierlitClient
from easierlit.errors import AppClosedError, RunFuncExecutionError
from easierlit.runtime import get_runtime


class _FakeUIMessage:
    sent: list[tuple[str, str | None]] = []

    def __init__(self, content: str, author: str | None = None):
        self.content = content
        self.author = author

    async def send(self):
        self.__class__.sent.append((self.content, self.author))
        return self


@pytest.fixture(autouse=True)
def _reset_runtime(monkeypatch):
    runtime = get_runtime()
    runtime.unbind()
    monkeypatch.setattr(chainlit_entry, "_APP_CLOSED_WARNING_EMITTED", False)
    monkeypatch.setattr(chainlit_entry, "_WORKER_FAILURE_UI_NOTIFIED", False)
    yield
    runtime.unbind()
    monkeypatch.setattr(chainlit_entry, "_APP_CLOSED_WARNING_EMITTED", False)
    monkeypatch.setattr(chainlit_entry, "_WORKER_FAILURE_UI_NOTIFIED", False)


def _patch_chainlit_context(monkeypatch):
    fake_session = SimpleNamespace(thread_id="thread-1", id="session-1")
    monkeypatch.setattr(
        chainlit_entry.cl,
        "context",
        SimpleNamespace(session=fake_session),
    )


def _sample_message():
    return SimpleNamespace(
        id="msg-1",
        content="/help",
        author="User",
        created_at=None,
        metadata=None,
    )


def test_on_message_swallows_closed_app_with_worker_error(monkeypatch, caplog):
    runtime = get_runtime()
    app = EasierlitApp()
    app.close()
    client = EasierlitClient(run_func=lambda _app: None, worker_mode="thread")
    runtime.bind(client=client, app=app)
    client._record_worker_error("Traceback (most recent call last):\nRuntimeError: boom")

    _FakeUIMessage.sent = []
    _patch_chainlit_context(monkeypatch)
    monkeypatch.setattr(chainlit_entry.cl, "Message", _FakeUIMessage)

    caplog.set_level("WARNING")
    asyncio.run(chainlit_entry._on_message(_sample_message()))
    asyncio.run(chainlit_entry._on_message(_sample_message()))

    assert len(_FakeUIMessage.sent) == 1
    assert "Server is shutting down" in _FakeUIMessage.sent[0][0]
    assert "server shutdown in progress" in caplog.text.lower()


def test_on_message_raises_when_closed_without_worker_error(monkeypatch):
    runtime = get_runtime()
    app = EasierlitApp()
    app.close()
    client = EasierlitClient(run_func=lambda _app: None, worker_mode="thread")
    runtime.bind(client=client, app=app)

    _patch_chainlit_context(monkeypatch)

    with pytest.raises(AppClosedError):
        asyncio.run(chainlit_entry._on_message(_sample_message()))


def test_on_app_shutdown_logs_summary_without_traceback(monkeypatch, caplog):
    runtime = get_runtime()
    app = EasierlitApp()
    client = EasierlitClient(run_func=lambda _app: None, worker_mode="thread")
    runtime.bind(client=client, app=app)

    worker_traceback = (
        "Traceback (most recent call last):\n"
        "RuntimeError: command '/new' failed: Chainlit context not found"
    )
    client._record_worker_error(worker_traceback)

    def fake_stop():
        raise RunFuncExecutionError(worker_traceback)

    monkeypatch.setattr(client, "stop", fake_stop)

    caplog.set_level("WARNING")
    asyncio.run(chainlit_entry._on_app_shutdown())

    assert "run_func crash acknowledged during shutdown" in caplog.text
    assert "command '/new' failed: Chainlit context not found" in caplog.text
    assert "Traceback (most recent call last)" not in caplog.text



class EasierlitError(Exception):
    """Base exception for Easierlit."""


class DataPersistenceNotEnabledError(EasierlitError):
    """Raised when Chainlit data persistence is required but not configured."""


class ThreadSessionNotActiveError(EasierlitError):
    """Raised when a thread has no active websocket session and no data layer fallback."""


class WorkerAlreadyRunningError(EasierlitError):
    """Raised when run() is called while the worker is already running."""


class WorkerNotRunningError(EasierlitError):
    """Raised when an operation expects a running worker but none exists."""


class RunFuncExecutionError(EasierlitError):
    """Raised when run_func crashes inside a worker."""


class AppClosedError(EasierlitError):
    """Raised when using EasierlitApp after close()."""



from easierlit.jwt_secret import ensure_jwt_secret


def test_ensure_jwt_secret_creates_secret_file(tmp_path):
    secret_path = tmp_path / ".chainlit" / "jwt.secret"

    secret = ensure_jwt_secret(secret_path=secret_path)

    assert secret_path.exists()
    assert secret_path.read_text(encoding="utf-8").strip() == secret
    assert len(secret.encode("utf-8")) >= 32


def test_ensure_jwt_secret_reuses_existing_valid_secret(tmp_path):
    secret_path = tmp_path / ".chainlit" / "jwt.secret"
    existing = "a" * 64
    secret_path.parent.mkdir(parents=True, exist_ok=True)
    secret_path.write_text(existing + "\n", encoding="utf-8")

    secret = ensure_jwt_secret(secret_path=secret_path)

    assert secret == existing
    assert secret_path.read_text(encoding="utf-8").strip() == existing


def test_ensure_jwt_secret_regenerates_short_secret(tmp_path):
    secret_path = tmp_path / ".chainlit" / "jwt.secret"
    secret_path.parent.mkdir(parents=True, exist_ok=True)
    secret_path.write_text("short\n", encoding="utf-8")

    secret = ensure_jwt_secret(secret_path=secret_path)

    assert secret != "short"
    assert len(secret.encode("utf-8")) >= 32
    assert secret_path.read_text(encoding="utf-8").strip() == secret


def test_ensure_jwt_secret_respects_custom_min_bytes(tmp_path):
    secret_path = tmp_path / ".chainlit" / "jwt.secret"

    secret = ensure_jwt_secret(secret_path=secret_path, min_bytes=80)

    assert len(secret.encode("utf-8")) >= 80



import pytest
from chainlit.data.utils import queue_until_user_message
from chainlit.types import PageInfo, PaginatedResponse

from easierlit import (
    DataPersistenceNotEnabledError,
    EasierlitApp,
    EasierlitAuthConfig,
    EasierlitClient,
    ThreadSessionNotActiveError,
)
from easierlit.runtime import get_runtime


class _FakeUser:
    def __init__(self, user_id: str):
        self.id = user_id


class FakeDataLayer:
    def __init__(self, users: dict[str, str] | None = None):
        self.created_steps = []
        self.updated_steps = []
        self.deleted_steps = []
        self.updated_threads = []
        self.deleted_threads = []
        self.requested_threads = []
        self.created_users = []
        self._users = users or {"known-user": "user-1"}
        self._threads = {"thread-1"}

    async def get_user(self, identifier: str):
        user_id = self._users.get(identifier)
        if user_id is None:
            return None
        return _FakeUser(user_id)

    async def create_user(self, user):
        user_id = f"user-created-{len(self.created_users) + 1}"
        self.created_users.append(
            {
                "id": user_id,
                "identifier": user.identifier,
                "metadata": user.metadata,
            }
        )
        self._users[user.identifier] = user_id
        return _FakeUser(user_id)

    async def list_threads(self, pagination, filters):
        return PaginatedResponse(
            pageInfo=PageInfo(hasNextPage=False, startCursor=None, endCursor=None),
            data=[
                {
                    "id": "thread-1",
                    "createdAt": "2024-01-01T00:00:00.000Z",
                    "name": "Thread 1",
                    "userId": "user-1",
                    "userIdentifier": "known-user",
                    "metadata": {},
                    "steps": [],
                    "elements": [],
                    "tags": [],
                }
            ],
        )

    async def get_thread(self, thread_id: str):
        self.requested_threads.append(thread_id)
        if thread_id not in self._threads:
            return None
        return {"id": thread_id, "name": "Thread 1", "tags": ["existing-tag"]}

    async def update_thread(
        self,
        thread_id: str,
        name=None,
        user_id=None,
        metadata=None,
        tags=None,
    ):
        self.updated_threads.append(
            {
                "thread_id": thread_id,
                "name": name,
                "user_id": user_id,
                "metadata": metadata,
                "tags": tags,
            }
        )
        self._threads.add(thread_id)

    async def delete_thread(self, thread_id: str):
        self.deleted_threads.append(thread_id)

    async def create_step(self, step_dict):
        self.created_steps.append(step_dict)

    async def update_step(self, step_dict):
        self.updated_steps.append(step_dict)

    async def delete_step(self, step_id: str):
        self.deleted_steps.append(step_id)


class FakeSQLiteDataLayer(FakeDataLayer):
    def __init__(self):
        super().__init__()
        self._conninfo = "sqlite+aiosqlite:///tmp/easierlit-test.db"

    async def list_threads(self, pagination, filters):
        return PaginatedResponse(
            pageInfo=PageInfo(hasNextPage=False, startCursor=None, endCursor=None),
            data=[
                {
                    "id": "thread-1",
                    "createdAt": "2024-01-01T00:00:00.000Z",
                    "name": "Thread 1",
                    "userId": "user-1",
                    "userIdentifier": "known-user",
                    "metadata": {},
                    "steps": [],
                    "elements": [],
                    "tags": '["tag-a", "tag-b"]',
                }
            ],
        )

    async def get_thread(self, thread_id: str):
        self.requested_threads.append(thread_id)
        if thread_id != "thread-1":
            return None
        return {"id": thread_id, "name": "Thread 1", "tags": '["tag-a"]'}


class _FakeEngineUrl:
    drivername = "sqlite+aiosqlite"


class _FakeEngine:
    url = _FakeEngineUrl()


class FakeSQLiteEngineOnlyDataLayer(FakeDataLayer):
    def __init__(self):
        super().__init__()
        self.engine = _FakeEngine()


class FakeDecoratedDataLayer(FakeDataLayer):
    @queue_until_user_message()
    async def create_step(self, step_dict):
        await super().create_step(step_dict)

    @queue_until_user_message()
    async def update_step(self, step_dict):
        await super().update_step(step_dict)

    @queue_until_user_message()
    async def delete_step(self, step_id: str):
        await super().delete_step(step_id)


@pytest.fixture(autouse=True)
def _reset_runtime():
    runtime = get_runtime()
    runtime.unbind()
    yield
    runtime.unbind()


def _apply_next_outgoing_command(app: EasierlitApp):
    runtime = get_runtime()
    command = app._pop_outgoing(timeout=1.0)
    runtime.run_coroutine_sync(runtime.apply_outgoing_command(command))
    return command


def test_thread_crud_requires_data_layer(monkeypatch):
    monkeypatch.setattr("easierlit.app.get_data_layer", lambda: None)
    app = EasierlitApp()

    with pytest.raises(DataPersistenceNotEnabledError):
        app.list_threads()


def test_thread_crud_with_data_layer(monkeypatch):
    fake = FakeDataLayer()
    monkeypatch.setattr("easierlit.app.get_data_layer", lambda: fake)

    app = EasierlitApp()
    threads = app.list_threads(first=5, user_identifier="known-user")
    assert len(threads.data) == 1

    thread = app.get_thread("thread-1")
    assert thread["id"] == "thread-1"
    assert thread["tags"] == ["existing-tag"]

    app.update_thread("thread-1", name="Renamed", metadata={"x": 1}, tags=["tag"])
    assert fake.updated_threads[0]["name"] == "Renamed"
    assert fake.updated_threads[0]["user_id"] is None
    assert fake.updated_threads[0]["tags"] == ["tag"]

    app.delete_thread("thread-1")
    assert fake.deleted_threads == ["thread-1"]


def test_history_preserves_original_step_order(monkeypatch):
    class _FakeDataLayerWithSteps(FakeDataLayer):
        async def get_thread(self, thread_id: str):
            self.requested_threads.append(thread_id)
            if thread_id != "thread-1":
                return None
            return {
                "id": thread_id,
                "name": "Thread 1",
                "tags": ["existing-tag"],
                "steps": [
                    {"id": "msg-1", "type": "user_message", "output": "hello"},
                    {"id": "msg-2", "type": "assistant_message", "output": "hi"},
                    {"id": "tool-1", "type": "tool", "output": '{"hits":2}'},
                    {"id": "run-1", "type": "run", "output": "done"},
                    "invalid-entry",
                ],
            }

    fake = _FakeDataLayerWithSteps()
    monkeypatch.setattr("easierlit.app.get_data_layer", lambda: fake)

    app = EasierlitApp()
    history = app.get_history("thread-1")

    assert history["thread"]["id"] == "thread-1"
    assert "steps" not in history["thread"]
    assert [item["id"] for item in history["items"]] == [
        "msg-1",
        "msg-2",
        "tool-1",
        "run-1",
    ]


def test_history_handles_missing_steps_key(monkeypatch):
    fake = FakeDataLayer()
    monkeypatch.setattr("easierlit.app.get_data_layer", lambda: fake)

    app = EasierlitApp()
    history = app.get_history("thread-1")

    assert history["thread"]["id"] == "thread-1"
    assert history["items"] == []


def test_new_thread_creates_when_missing(monkeypatch):
    fake = FakeDataLayer()
    monkeypatch.setattr("easierlit.app.get_data_layer", lambda: fake)
    monkeypatch.setattr("easierlit.app.uuid4", lambda: "thread-new")

    app = EasierlitApp()
    thread_id = app.new_thread(name="Created", metadata={"x": 1}, tags=["tag"])

    assert thread_id == "thread-new"
    assert fake.updated_threads[0]["thread_id"] == thread_id
    assert fake.updated_threads[0]["name"] == "Created"
    assert fake.updated_threads[0]["metadata"] == {"x": 1}
    assert fake.updated_threads[0]["tags"] == ["tag"]


def test_new_thread_retries_when_generated_id_exists(monkeypatch):
    fake = FakeDataLayer()
    fake._threads.add("thread-collision")
    monkeypatch.setattr("easierlit.app.get_data_layer", lambda: fake)

    generated_ids = iter(["thread-collision", "thread-created"])
    monkeypatch.setattr("easierlit.app.uuid4", lambda: next(generated_ids))

    app = EasierlitApp()
    thread_id = app.new_thread(name="Created")

    assert thread_id == "thread-created"
    assert fake.updated_threads[0]["thread_id"] == "thread-created"
    assert fake.requested_threads[:2] == ["thread-collision", "thread-created"]


def test_new_thread_raises_when_unique_id_allocation_fails(monkeypatch):
    fake = FakeDataLayer()
    fake._threads.add("thread-duplicate")
    monkeypatch.setattr("easierlit.app.get_data_layer", lambda: fake)
    monkeypatch.setattr("easierlit.app.uuid4", lambda: "thread-duplicate")

    app = EasierlitApp()
    with pytest.raises(RuntimeError, match="Failed to allocate unique thread_id"):
        app.new_thread(name="Duplicate")

    assert len(fake.requested_threads) == 16
    assert fake.updated_threads == []


def test_update_thread_raises_when_thread_missing(monkeypatch):
    fake = FakeDataLayer()
    monkeypatch.setattr("easierlit.app.get_data_layer", lambda: fake)

    app = EasierlitApp()
    with pytest.raises(ValueError, match="not found"):
        app.update_thread("missing-thread", name="Renamed")


def test_sqlite_update_thread_serializes_tags(monkeypatch):
    fake = FakeSQLiteDataLayer()
    monkeypatch.setattr("easierlit.app.get_data_layer", lambda: fake)

    app = EasierlitApp()
    app.update_thread("thread-1", tags=["run-func-created"])

    assert fake.updated_threads[0]["user_id"] is None
    assert fake.updated_threads[0]["tags"] == '["run-func-created"]'


def test_sqlite_update_thread_serializes_tags_with_engine_drivername(monkeypatch):
    fake = FakeSQLiteEngineOnlyDataLayer()
    monkeypatch.setattr("easierlit.app.get_data_layer", lambda: fake)

    app = EasierlitApp()
    app.update_thread("thread-1", tags=["run-func-created"])

    assert fake.updated_threads[0]["user_id"] is None
    assert fake.updated_threads[0]["tags"] == '["run-func-created"]'


def test_update_thread_auto_sets_owner_from_auth_existing_user(monkeypatch):
    fake = FakeDataLayer(users={"admin": "user-admin"})
    monkeypatch.setattr("easierlit.app.get_data_layer", lambda: fake)

    runtime = get_runtime()
    app = EasierlitApp()
    client = EasierlitClient(run_func=lambda _app: None)
    runtime.bind(
        client=client,
        app=app,
        auth=EasierlitAuthConfig(
            username="admin",
            password="admin",
            identifier="admin",
            metadata={"role": "admin"},
        ),
    )

    app.update_thread("thread-1", name="Owned")

    assert fake.updated_threads[0]["thread_id"] == "thread-1"
    assert fake.updated_threads[0]["user_id"] == "user-admin"
    assert fake.created_users == []


def test_update_thread_auto_creates_owner_when_missing(monkeypatch):
    fake = FakeDataLayer(users={})
    monkeypatch.setattr("easierlit.app.get_data_layer", lambda: fake)

    runtime = get_runtime()
    app = EasierlitApp()
    client = EasierlitClient(run_func=lambda _app: None)
    runtime.bind(
        client=client,
        app=app,
        auth=EasierlitAuthConfig(
            username="admin",
            password="admin",
            identifier="admin",
            metadata={"role": "admin"},
        ),
    )

    app.update_thread("thread-1", name="Owned")

    assert fake.created_users[0]["identifier"] == "admin"
    assert fake.created_users[0]["metadata"] == {"role": "admin"}
    assert fake.updated_threads[0]["user_id"] == "user-created-1"


def test_sqlite_get_thread_normalizes_tags(monkeypatch):
    fake = FakeSQLiteDataLayer()
    monkeypatch.setattr("easierlit.app.get_data_layer", lambda: fake)

    app = EasierlitApp()
    thread = app.get_thread("thread-1")

    assert thread["tags"] == ["tag-a"]


def test_sqlite_list_threads_normalizes_tags(monkeypatch):
    fake = FakeSQLiteDataLayer()
    monkeypatch.setattr("easierlit.app.get_data_layer", lambda: fake)

    app = EasierlitApp()
    threads = app.list_threads(first=5)

    assert threads.data[0]["tags"] == ["tag-a", "tag-b"]


def test_message_crud_falls_back_to_data_layer_when_no_session(monkeypatch):
    fake = FakeDataLayer()
    monkeypatch.setattr("easierlit.runtime.get_data_layer", lambda: fake)

    app = EasierlitApp()
    message_id = app.add_message("thread-1", "hello", author="Bot")
    _apply_next_outgoing_command(app)
    assert fake.created_steps[0]["id"] == message_id
    assert fake.created_steps[0]["type"] == "assistant_message"
    assert fake.created_steps[0]["name"] == "Bot"

    app.update_message("thread-1", message_id, "updated")
    _apply_next_outgoing_command(app)
    assert fake.updated_steps[0]["id"] == message_id
    assert fake.updated_steps[0]["type"] == "assistant_message"

    app.delete_message("thread-1", message_id)
    _apply_next_outgoing_command(app)
    assert fake.deleted_steps == [message_id]


def test_message_crud_raises_without_session_and_data_layer(monkeypatch):
    monkeypatch.setattr("easierlit.runtime.get_data_layer", lambda: None)
    app = EasierlitApp()

    app.add_message("thread-unknown", "hello")
    with pytest.raises(ThreadSessionNotActiveError):
        _apply_next_outgoing_command(app)


def test_message_crud_fallback_works_with_queue_decorated_data_layer(monkeypatch):
    from types import SimpleNamespace

    from chainlit.context import context_var

    fake = FakeDecoratedDataLayer()
    context_calls: list[tuple[str, str]] = []

    def fake_init_http_context(*, thread_id: str, client_type: str):
        context_calls.append((thread_id, client_type))
        context_var.set(SimpleNamespace(session=object()))

    monkeypatch.setattr("easierlit.runtime.get_data_layer", lambda: fake)
    monkeypatch.setattr("easierlit.runtime.init_http_context", fake_init_http_context)

    app = EasierlitApp()
    message_id = app.add_message("thread-queue", "hello", author="Bot")
    _apply_next_outgoing_command(app)

    app.update_message("thread-queue", message_id, "updated")
    _apply_next_outgoing_command(app)

    app.delete_message("thread-queue", message_id)
    _apply_next_outgoing_command(app)

    assert len(context_calls) == 3
    assert context_calls[0] == ("thread-queue", "webapp")
    assert fake.created_steps[0]["id"] == message_id
    assert fake.created_steps[0]["type"] == "assistant_message"
    assert fake.updated_steps[0]["id"] == message_id
    assert fake.updated_steps[0]["type"] == "assistant_message"
    assert fake.deleted_steps == [message_id]


def test_tool_and_thought_crud_falls_back_to_data_layer_when_no_session(monkeypatch):
    fake = FakeDataLayer()
    monkeypatch.setattr("easierlit.runtime.get_data_layer", lambda: fake)

    app = EasierlitApp()

    tool_message_id = app.add_tool(
        thread_id="thread-1",
        tool_name="SearchTool",
        content='{"query":"books"}',
    )
    _apply_next_outgoing_command(app)
    assert fake.created_steps[0]["id"] == tool_message_id
    assert fake.created_steps[0]["type"] == "tool"
    assert fake.created_steps[0]["name"] == "SearchTool"

    app.update_tool(
        thread_id="thread-1",
        message_id=tool_message_id,
        tool_name="SearchTool",
        content='{"results":2}',
    )
    _apply_next_outgoing_command(app)
    assert fake.updated_steps[0]["id"] == tool_message_id
    assert fake.updated_steps[0]["type"] == "tool"
    assert fake.updated_steps[0]["name"] == "SearchTool"

    thought_message_id = app.add_thought(
        thread_id="thread-1",
        content="Need one more retrieval pass.",
    )
    _apply_next_outgoing_command(app)
    assert fake.created_steps[1]["id"] == thought_message_id
    assert fake.created_steps[1]["type"] == "tool"
    assert fake.created_steps[1]["name"] == "Reasoning"

    app.update_thought(
        thread_id="thread-1",
        message_id=thought_message_id,
        content="Enough context collected.",
    )
    _apply_next_outgoing_command(app)
    assert fake.updated_steps[1]["id"] == thought_message_id
    assert fake.updated_steps[1]["type"] == "tool"
    assert fake.updated_steps[1]["name"] == "Reasoning"

    app.delete_message("thread-1", tool_message_id)
    _apply_next_outgoing_command(app)
    app.delete_message("thread-1", thought_message_id)
    _apply_next_outgoing_command(app)
    assert fake.deleted_steps == [tool_message_id, thought_message_id]
